from faker.providers.address.pt_BR import Provider as AddressProvider


class CepProvider(AddressProvider):
    namespace = 'address'  # NAMESPACE SHOULD BE ADDRESS

    def postcode(self, mask: bool = True):
        if mask:
            return self.bothify('#####-###')
        else:
            return self.bothify('########')



from io import StringIO

from src.lib.formatters.csv import CsvFormatter
from src.tests.lib.formatters.fixtures import *  # noqa: F403, F401


class TestCsvFormatter(object):
    """Unit-test of CsvWriter class"""

    def test_writting_dataframe_with_records(self,
                                             pandas_dataframe_with_data):
        buffer = StringIO()
        csv_writer = CsvFormatter(specification={})
        csv_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=buffer)

        assert isinstance(buffer.getvalue(), str) is True
        assert len(buffer.getvalue()) > 0

    def test_writting_dataframe_without_records(self,
                                                pandas_dataframe_without_data):
        buffer = StringIO()
        csv_writer = CsvFormatter(specification={})
        csv_writer.format(dataframe=pandas_dataframe_without_data,
                          path_or_buffer=buffer)

        assert isinstance(buffer.getvalue(), str) is True
        assert len(buffer.getvalue()) == 0

    def test_date_format_epoch(self, dataframe_with_datetime):
        buffer = StringIO()
        spec = {'options': {'date_format': 'epoch'}}
        csv_writer = CsvFormatter(specification=spec)
        csv_writer.format(dataframe=dataframe_with_datetime,
                          path_or_buffer=buffer)

        csv_text = buffer.getvalue()
        assert isinstance(csv_text, str) is True
        assert len(csv_text) > 0

        for date_ in dataframe_with_datetime['at'].tolist():
            assert str(date_) in csv_text

    def test_date_format_iso(self, dataframe_with_datetime):
        buffer = StringIO()
        spec = {'options': {'date_format': 'iso'}}
        csv_writer = CsvFormatter(specification=spec)
        csv_writer.format(dataframe=dataframe_with_datetime,
                          path_or_buffer=buffer)

        csv_text = buffer.getvalue()
        assert isinstance(csv_text, str) is True
        assert len(csv_text) > 0

        for date_ in dataframe_with_datetime['at'].tolist():
            assert date_.isoformat() in csv_text



import csv

from datetime import datetime
from pandas import DataFrame


class CsvFormatter(object):
    """Class that receive pandas dataframe
    and write it down in CSV format
    """
    key = 'csv'

    def __init__(self, specification):
        self.default = {'index': False, 'sep': ','}
        self.specification = specification

    @staticmethod
    def rules():
        return {'required': {},
                'optional': {
                    'options.header': {'none': False, 'type': bool},
                    'options.sep': {'none': False, 'type': str},
                    'options.index': {'none': False, 'type': bool}
                }}

    def format(self, dataframe: DataFrame, path_or_buffer) -> None:
        """Format dataframe to csv.

        Parameters:
         - dataframe - pandas.DataFrame: dataframe containing the records.
        """
        parameters = self.default
        options = self.specification.get('options', {})
        parameters.update(options)

        date_format = options.get('date_format')
        if date_format == 'epoch':
            parameters.pop('date_format')
            epoch = datetime(1970, 1, 1)
            for column in dataframe.columns:
                if dataframe[column].dtype == 'datetime64[ns]':
                    dataframe[column] = \
                        dataframe[column].apply(lambda x: int((x - epoch)
                                                .total_seconds()))
        elif date_format == 'iso':
            parameters.update({'date_format': '%Y-%m-%dT%H:%M:%SZ'})

        if dataframe.shape[0] > 0:
            return dataframe.to_csv(path_or_buf=path_or_buffer,
                                    quoting=csv.QUOTE_NONNUMERIC, **parameters)



from faker.providers.ssn.pt_BR import Provider as SsnProvider


class DocumentProvider(SsnProvider):
    namespace = 'ssn'

    def titulo_eleitoral(self, mask: bool = True):
        if mask:
            return self.bothify('#### #### ####')
        else:
            return self.bothify('############')

    def cnh(self, mask: bool = True):
        if mask:
            return self.bothify('### ### ### ##')
        else:
            return self.bothify('###########')



import json
import requests
import os

from io import StringIO
from pandas import DataFrame


class S3PresignedUrlRemoteWriter(object):
    key = 's3-url'
    file_path = None

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification

    def before_write(self):
        options = self.specification.get('options')

        if options is None:
            file_path = input("Enter the signed url json file path: ")
            if os.path.isfile(file_path) is False:
                raise ValueError('File not found')
            self.file_path = file_path

    def write(self, dataframe: DataFrame):
        with open(self.file_path, 'r') as file:
            singed_post = json.loads(file.read())

        signed_url = singed_post.get('url')
        fields = singed_post.get('fields')

        buffer = StringIO()
        self.formatter.format(dataframe=dataframe,
                              path_or_buffer=buffer)

        files = {'file': buffer.getvalue()}

        try:
            response = requests.post(signed_url,
                                     data=fields,
                                     files=files)
            response.raise_for_status()
        except Exception as e:
            raise requests.RequestException("Can't send data to this given"
                                            "URI try to check if still valid",
                                            e)
        return signed_url

    @staticmethod
    def rules():
        return {'required': {},
                'optional': {'options.path': {'none': False, 'type': str}}}



from src.generators.datatypes.faker_custom.address.pt_BR import CepProvider
from src.generators.datatypes.faker_custom.ssn.pt_BR import DocumentProvider
from src.generators.datatypes.faker_custom.company.pt_BR import CompanyProvider

custom_providers = {
    'pt_BR': [DocumentProvider, CepProvider, CompanyProvider]
}



from src.lib.writers.remotes.azure_bs import AzureBSRemoteWriter
from src.lib.formatters import formatters
from src.tests.lib.writers.remotes.fixtures import *  # noqa: F403, F401
import pytest
from io import StringIO
from csv import reader
from azure.storage.blob import BlobServiceClient


class TestAzureBSWriter(object):
    def test_azure_without_connection_string(self,
                                             mocker,
                                             specification_azure_bs):
        csv_formatter = formatters['csv']({"options": {"header": True}})
        mock = mocker.patch.object(AzureBSRemoteWriter, 'get_credentials')
        mock.return_value = ""

        with pytest.raises(Exception):
            AzureBSRemoteWriter(formatter=csv_formatter,
                                specification=specification_azure_bs)

    def test_write_dataframe_csv(self, mocker,
                                 pandas_dataframe_with_data,
                                 specification_azure_bs):
        mock = mocker.patch.object(BlobServiceClient, 'from_connection_string')
        mock.return_value = None

        mock_credential = mocker.patch.object(
            AzureBSRemoteWriter, 'get_credentials')
        mock_credential.return_value = "MyCredential"

        buffer = StringIO()
        mock_buffer = mocker.patch.object(AzureBSRemoteWriter, 'open_buffer')
        mock_buffer.return_value = buffer

        mock_upload = mocker.patch.object(
            AzureBSRemoteWriter, 'upload_blob')
        mock_upload.return_value = None

        csv_formatter = formatters['csv']({"options": {"header": True}})
        azure_bs = AzureBSRemoteWriter(formatter=csv_formatter,
                                       specification=specification_azure_bs)
        azure_bs.write(pandas_dataframe_with_data)
        expected = pandas_dataframe_with_data.to_dict(orient="list")

        buffer.seek(0)
        csv_reader = reader(buffer.readlines(), delimiter=',')
        csv_reader = [x for x in csv_reader]
        headers = csv_reader[0]
        rows = csv_reader[1:len(csv_reader)]
        csv_reader = {headers[index]: [row[index]
                                       for row in rows]
                      for index in range(0, len(headers))}

        assert csv_reader == expected

    def test_write_dataframe_json(self, mocker,
                                  pandas_dataframe_with_data,
                                  specification_azure_bs):
        mock = mocker.patch.object(BlobServiceClient, 'from_connection_string')
        mock.return_value.open.return_value = None

        mock_credential = mocker.patch.object(
            AzureBSRemoteWriter, 'get_credentials')
        mock_credential.return_value = "MyCredential"

        buffer = StringIO()
        mock_buffer = mocker.patch.object(AzureBSRemoteWriter, 'open_buffer')
        mock_buffer.return_value = buffer

        mock_upload = mocker.patch.object(
            AzureBSRemoteWriter, 'upload_blob')
        mock_upload.return_value = None

        json_formatter = formatters['json-array'](
            {"options": {"orient": "records"}})

        azure_bs = AzureBSRemoteWriter(formatter=json_formatter,
                                       specification=specification_azure_bs)
        azure_bs.write(pandas_dataframe_with_data)
        expected = pandas_dataframe_with_data.to_json(orient="records")

        assert buffer.getvalue() == expected



from getpass import getpass


class DatabaseWriter(object):
    @staticmethod
    def rules():
        return {
            'required': {
                'options.engine': {'none': False, 'type': str},
                'options.host': {'none': False, 'type': str},
                'options.database': {'none': False, 'type': str},
                'options.username': {'none': False, 'type': str},
            },
            'optional': {
                'options.schema': {'none': False, 'type': str},
                'options.method': {'none': False, 'type': str},
                'options.port': {'none': False, 'type': int},
            }
        }

    def before_write(self):
        host = self.specification \
                   .get('options') \
                   .get('host')
        user = self.specification \
                   .get('options') \
                   .get('username')
        print(f"Database host: {host}")
        print(f"Username: {user}")
        self.specification['options']['password'] = \
            getpass(f"Type {user} password: ")



from faker import Faker
import inspect
from src.generators.datatypes.faker_custom import custom_providers


class FakerProxy(object):
    def __init__(self, locale: str):
        self.locale = locale
        self.faker = Faker(locale=self.locale)
        self.add_custom_providers()
        self.blacklist_namespaces = ['generic']
        self.blacklist_generator = ['time_series']
        self.list_of_generators = self.__generators_types()
        self.__infer_types()

    def add_custom_providers(self):
        for custom_locale, providers in custom_providers.items():
            if custom_locale == self.locale or custom_locale == 'global':
                for provider in providers:
                    self.faker.add_provider(provider)

    def __generators_types(self) -> list:
        def is_not_namespace_blacklist(generator):
            namespace = self.__get_namespace(generator)
            return namespace not in self.blacklist_namespaces

        def is_not_generator_blacklist(generator):
            return generator.__name__ not in self.blacklist_generator

        return [generator for _, generator
                in self.faker.factories[0].__dict__.items()
                if callable(generator) and
                is_not_generator_blacklist(generator) and
                is_not_namespace_blacklist(generator)]

    def __infer_types(self) -> None:
        for generator_type in self.list_of_generators:
            namespace = self.__get_namespace(generator_type)
            self.__infer_type(generator_type=generator_type,
                              namespace=namespace)

    def __infer_type(self, generator_type, namespace) -> None:
        self.__setattr__(generator_type.__name__,
                         FakerType(generator_type=generator_type,
                                   namespace=namespace))

    def __get_type_info(self, generator_type):
        return inspect.getfullargspec(generator_type)

    def __get_namespace(self, obj) -> str:
        try:
            package = inspect.getmodule(obj).__name__.split('.')
            if 'faker_custom' in package:
                return package[-2]
            return package[2]
        except IndexError:
            return "generic"


class FakerType(object):
    optional_arguments = True

    def __init__(self, generator_type, namespace):
        self.generator_type = generator_type
        self.key = generator_type.__name__
        self.namespace = namespace

    def __call__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

    def generate(self):
        return self.__parse_data(self.generator_type(*self.args,
                                                     **self.kwargs))

    def generate_records(self, num_of_records, progress=None):
        results = []
        for _ in range(num_of_records):
            results.append(self.generate())
            progress and progress()

        return results

    def __parse_data(self, data):
        if isinstance(data, str):
            data = self.__remove_break_lines(data)
        return data

    def __remove_break_lines(self, data: str):
        return data.replace('\n', ' ')

    @staticmethod
    def rules():
        return {'required': {},
                'optional': {}}

    def sample(self):
        return self.__parse_data(self.generator_type())



import json
import pytest

from mock import mock_open, patch
from os import remove
from pandas import DataFrame
from pytest import raises

from src.generators.handler import GeneratorsHandler
from src.tests.generators.handler_fixtures import *  # noqa: F403, F401
from src.lib.writers import writers


class TestGeneratorsHandler(object):
    """Unit-test of GeneratorsHandler class"""

    def test_valid_specification(self, mocker, valid_specification):
        data = json.dumps(valid_specification)
        with patch('builtins.open', mock_open(read_data=data)) as mock:
            GeneratorsHandler(arguments={'config_file': None})

            mock.assert_called()

    def test_without_dataset(self, mocker, no_datasets_specification):
        data = json.dumps(no_datasets_specification)
        with patch('builtins.open', mock_open(read_data=data)) as mock:

            with raises(ValueError):
                GeneratorsHandler(arguments={'config_file': None})
            mock.assert_called()

    def test_invalid_dataset(self, mocker, invalid_dataset_specification):
        data = json.dumps(invalid_dataset_specification)
        with patch('builtins.open', mock_open(read_data=data)) as mock:
            with raises(ValueError):
                GeneratorsHandler(arguments={'config_file': None})
            mock.assert_called()

    def test_generate(self, mocker, valid_specification):
        mock_write = mocker.patch \
                           .object(GeneratorsHandler, 'write_dataframe')
        mock_write.return_value = 'file_path'
        mock_config = mocker.patch \
                            .object(GeneratorsHandler,
                                    'valid_specification_dataset')
        mock_config.return_value = valid_specification

        handler = GeneratorsHandler(arguments={'config_file': None})
        for key, format_, path in handler.generate():
            assert key is not None
            assert isinstance(key, str)
            assert format_ is not None
            assert isinstance(format_, str)
            assert path is not None
            assert isinstance(path, str)

        mock_write.assert_called()

    def test_write(self, mocker, valid_specification):
        mock_config = mocker.patch \
                            .object(GeneratorsHandler,
                                    'valid_specification_dataset')
        mock_config.return_value = valid_specification

        handler = GeneratorsHandler(arguments={'config_file': None})
        for key in valid_specification['datasets'].keys():
            dataset = valid_specification['datasets'][key]
            dataset_format = dataset['format']
            dataframe = handler.generate_dataframe(dataset)
            for destination in dataset['serializers']['to']:
                file_type = destination['type']
                writer = writers[file_type]
                mock = mocker.patch.object(writer, 'write')
                mock.return_value = 'file_path'
                file_path = handler.write_dataframe(dataframe=dataframe,
                                                    destination=destination,
                                                    format_=dataset_format)

                mock.assert_called()
                assert file_path == 'file_path'

    def test_simple_specification_dataframe(self,
                                            mocker,
                                            simple_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = simple_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == simple_specification['size']
        for field in simple_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_argumented_specification_dataframe(self,
                                                mocker,
                                                argumented_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = argumented_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == argumented_specification['size']
        for field in argumented_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_integer_dataframe(self, mocker, integer_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = integer_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == integer_specification['size']
        for field in integer_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_bool_dataframe(self, mocker, bool_specification):
        with patch('builtins.open',
                   mock_open(read_data=json.dumps(bool_specification))):
            dataset = bool_specification['datasets']['sample']
            expected_size = dataset['size']
            expected_fields = dataset['fields']

            handler = GeneratorsHandler({"config_file": None})
            dataframe = handler.generate_dataframe(dataset)

            assert isinstance(dataframe, DataFrame) is True
            assert dataframe.shape[0] == expected_size
            for field in expected_fields:
                assert dataframe[field['name']].dtype.name == field['expected']

    def test_name_dataframe(self, mocker, name_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = name_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == name_specification['size']
        for field in name_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_char_dataframe(self, mocker, char_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = char_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == char_specification['size']
        for field in char_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_float_dataframe(self, mocker, float_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = float_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == float_specification['size']
        for field in float_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_integer_sequence_dataframe(self,
                                        mocker,
                                        integer_sequence_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = integer_sequence_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == integer_sequence_specification['size']
        for field in integer_sequence_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_timestamp_sequence_dataframe(self,
                                          mocker,
                                          timestamp_sequence_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = timestamp_sequence_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == timestamp_sequence_specification['size']
        for field in timestamp_sequence_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_string_dataframe(self, mocker, string_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = string_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == string_specification['size']
        for field in string_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_timestamp_dataframe(self, mocker, timestamp_specification):
        mock = mocker.patch \
                     .object(GeneratorsHandler, 'valid_specification_dataset')
        mock.return_value = timestamp_specification
        handler = GeneratorsHandler({"config_file": None})
        specification = handler.valid_specification_dataset()
        dataframe = handler.generate_dataframe(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == timestamp_specification['size']
        for field in timestamp_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_valid_spec_handler(self, mocker, valid_specification):
        from os.path import abspath
        from uuid import uuid4

        absolute_path = abspath(".")
        file_name = uuid4()
        with open(f"{absolute_path}/{file_name}", "w") as f:
            json.dump(valid_specification, f)

        handler = \
            GeneratorsHandler({"config_file": f"{absolute_path}/{file_name}"})

        remove(f"{absolute_path}/{file_name}")
        assert handler

    def test_invalid_no_ids_spec_handler(self, mocker,
                                         invalid_no_ids_dataset):
        with patch('builtins.open',
                   mock_open(read_data=json
                             .dumps(invalid_no_ids_dataset))) as mock:
            with pytest.raises(ValueError):
                GeneratorsHandler({"config_file": ""})
            mock.assert_called()

    def test_invalid_no_dataset_spec_handler(self, mocker,
                                             no_datasets_specification):
        from os.path import abspath
        from uuid import uuid4

        absolute_path = abspath(".")
        file_name = uuid4()
        with open(f"{absolute_path}/{file_name}", "w") as f:
            json.dump(no_datasets_specification, f)

        with pytest.raises(ValueError):
            GeneratorsHandler({"config_file": f"{absolute_path}/{file_name}"})
        remove(f"{absolute_path}/{file_name}")

    def test_invalid_no_dataset_info(self, mocker,
                                     invalid_dataset_specification):
        from os.path import abspath
        from uuid import uuid4

        absolute_path = abspath(".")
        file_name = uuid4()
        with open(f"{absolute_path}/{file_name}", "w") as f:
            json.dump(invalid_dataset_specification, f)

        with pytest.raises(ValueError):
            GeneratorsHandler({"config_file": f"{absolute_path}/{file_name}"})
        remove(f"{absolute_path}/{file_name}")



from pytest import raises

from src.lib.config.fields import FieldsConfiguration
from src.lib.config.inspector import ConfigurationInpector
from src.tests.lib.config.fixtures import *  # noqa: F403, F401


class TestFieldsConfiguration(object):
    """Unit-test of TestFieldsConfiguration class"""

    def test_valid_csv_specification(self, valid_csv_specification):
        inspector = ConfigurationInpector()

        dataset = valid_csv_specification['datasets']['valid']
        fields = dataset['fields']

        for field in fields:
            rules = FieldsConfiguration.get_rules(field, dataset['locale'])
            required = rules['required']
            inspect = inspector. \
                inspect_rules(rules=required,
                              configuration=field)
            assert inspect is None

    def test_invalid_csv_configuration(self, invalid_csv_specification):
        inspector = ConfigurationInpector()
        dataset = invalid_csv_specification['datasets']['invalid']
        fields = dataset['fields']

        for field in fields:
            with raises(ValueError):
                rules = FieldsConfiguration.get_rules(field, dataset['locale'])
                required = rules['required']

                inspector. \
                    inspect_rules(rules=required,
                                  configuration=field)



from faker.providers.company.pt_BR import Provider


class CompanyProvider(Provider):
    def cnpj(self, mask: bool = True):
        digits = self.company_id()
        if mask:
            return '{}.{}.{}/{}-{}'.format(digits[:2], digits[2:5],
                                           digits[5:8], digits[8:12],
                                           digits[12:])
        else:
            return self.company_id()



import random

from datetime import datetime
from pytest import raises

from src.generators.datatypes.sequence_timestamp import TimestampSequenceType
from src.tests.generators.datatypes.generators_fixtures import *  # noqa: F403, F401, E501

TEST_NUM_OF_RECORDS = random.randint(1, 1000)


class TestTimestampGenerator:

    def test_num_of_records(self, generate):
        records = generate(TimestampSequenceType,
                           TEST_NUM_OF_RECORDS,
                           start_at="2019-01-01T01:00:00Z",)
        assert len(records) == TEST_NUM_OF_RECORDS

    def test_all_records_datetime(self, generate):
        records = generate(TimestampSequenceType,
                           TEST_NUM_OF_RECORDS,
                           start_at="2019-01-01T01:00:00Z",)
        for record in records:
            assert isinstance(record, datetime)

    def test_not_receive_date_format(self, generate):
        with raises(ValueError):
            generate(TimestampSequenceType,
                     TEST_NUM_OF_RECORDS,
                     start_at="sfsdfsdfs")

    def test_minutes_sequence_part(self, generate):
        dateparts = ['second', 'minute', 'hour', 'day', 'month', 'year']
        for datepart in dateparts:
            records = generate(TimestampSequenceType,
                               TEST_NUM_OF_RECORDS,
                               start_at="2019-01-01T01:00:00Z",
                               datepart=datepart)
            assert len(records) == TEST_NUM_OF_RECORDS



from typing import List


class SequenceType:
    key = 'integer:sequence'
    namespace = 'basic_type'
    optional_arguments = True

    def __init__(self, start_at=0, step=1, *args, **kwargs):
        self.start_at = start_at
        self.step = step

    @staticmethod
    def rules():
        def validate(value):
            min_ = -2147483648
            max_ = 2147483648
            if value < min_ or value > max_:
                raise ValueError("Sequence `start_at` generator must be "
                                 f"greater than or equals {min_} and "
                                 f"less than or equals {max_}")
        return {'required': {'generator.start_at': {'none': False,
                                                    'type': int,
                                                    'custom': [validate]}},
                'optional': {}}

    def generate(self, num_of_records: int = 5) -> List[int]:
        return self.generate_records(num_of_records=num_of_records)

    def generate_records(self, num_of_records, progress=None) -> List[int]:
        return self.__generate_sequence(num_of_rows=num_of_records,
                                        step=self.step,
                                        progress=progress)

    def __generate_sequence(self,
                            num_of_rows,
                            step,
                            progress=None) -> List[int]:
        results = []
        for x in range(self.start_at,
                       self.start_at + num_of_rows * step,
                       step):
            results.append(x)
            progress and progress()
        return results



from pytest import raises
from src.lib.formatters.sql import SQLFormatter
from src.tests.lib.formatters.fixtures import *  # noqa: F403, F401
from io import StringIO


class TestSqlFormatter(object):
    """Unit-test of SQLWriter class"""

    def test_writting_dataframe_with_records(self,
                                             sql_specification_format,
                                             pandas_dataframe_with_data):
        buffer = StringIO()
        sql_writer = SQLFormatter(
            specification=sql_specification_format('mytable', mode='append'))
        sql_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=buffer)

        assert isinstance(buffer.getvalue(), str) is True
        assert len(buffer.getvalue()) > 0

    def test_sql_script_is_append(self,
                                  sql_specification_format,
                                  pandas_dataframe_with_data):
        buffer = StringIO()
        sql_writer = SQLFormatter(
            specification=sql_specification_format('mytable', mode='append'))
        sql_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=buffer)

        assert "INSERT INTO".lower() in buffer.getvalue().lower()

    def test_sql_script_contain_dataframe_columns(self,
                                                  sql_specification_format,
                                                  pandas_dataframe_with_data):
        buffer = StringIO()
        sql_writer = SQLFormatter(
            specification=sql_specification_format('mytable', mode='append'))
        sql_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=buffer)

        for column in pandas_dataframe_with_data:
            assert column.lower() in buffer.getvalue().lower()

    def test_sql_script_contain_dataframe_rows(self,
                                               sql_specification_format,
                                               pandas_dataframe_with_data):
        buffer = StringIO()
        sql_writer = SQLFormatter(
            specification=sql_specification_format('mytable', mode='append'))
        sql_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=buffer)

        for index, row in pandas_dataframe_with_data.iterrows():
            for column in pandas_dataframe_with_data.columns:
                assert row[column].lower() in buffer.getvalue().lower()

    def test_sql_script_contain_rows_and_index(self,
                                               sql_specification_format,
                                               pandas_dataframe_with_data):
        buffer = StringIO()
        index_flag = True
        index_label = 'myindexlabel'
        csv_writer = SQLFormatter(
            specification=sql_specification_format('mytable',
                                                   index=index_flag,
                                                   index_label=index_label,
                                                   mode='append'))
        csv_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=buffer)

        assert index_label in buffer.getvalue().lower()
        for index, row in pandas_dataframe_with_data.iterrows():
            for column in pandas_dataframe_with_data.columns:
                assert row[column].lower() in buffer.getvalue().lower()
                assert str(index) in buffer.getvalue().lower()

    def test_writting_dataframe_without_records(self,
                                                pandas_dataframe_without_data):
        buffer = StringIO()
        sql_writer = SQLFormatter(specification={
            'options': {
                'table_name': 'mytable',
                'batch_size': 2,
                'mode': 'append',
                'schema': {}
            }
        })
        sql_writer.format(dataframe=pandas_dataframe_without_data,
                          path_or_buffer=buffer)

        assert isinstance(buffer.getvalue(), str) is True
        assert len(buffer.getvalue()) == 0

    def test_truncate(self, sql_specification_format,
                      pandas_dataframe_with_data):
        buffer = StringIO()
        table_name = 'mytable'
        sql_writer = SQLFormatter(
            specification=sql_specification_format('mytable', mode='truncate'))
        sql_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=buffer)

        text = buffer.getvalue()
        assert isinstance(text, str) is True
        assert len(text) > 0
        assert f'TRUNCATE {table_name}' in text

    def test_replace(self,
                     replace_dataframe,
                     fixture_spec_replace):
        buffer = StringIO()
        sql_writer = SQLFormatter(specification=fixture_spec_replace)
        sql_writer.format(replace_dataframe, buffer)

        table_name = 'My_table'
        text = buffer.getvalue()
        assert isinstance(text, str) is True
        assert len(text) > 0
        assert f'DROP TABLE IF EXISTS {table_name};' in text

    def test_replace_without_schema(self,
                                    replace_dataframe,
                                    fixture_spec_replace_without_schema):
        buffer = StringIO()
        sql_writer = SQLFormatter(
            specification=fixture_spec_replace_without_schema)
        with raises(ValueError):
            sql_writer.format(replace_dataframe, buffer)

    def test_invalid_mode(self, sql_specification_format,
                          pandas_dataframe_with_data):
        buffer = StringIO()
        table_name = 'mytable'
        sql_writer = SQLFormatter(
            specification=sql_specification_format(table_name, mode=''))

        with raises(ValueError):
            sql_writer.format(dataframe=pandas_dataframe_with_data,
                              path_or_buffer=buffer)

    def test_sql_script_is_append2(self,
                                   replace_dataframe,
                                   fixture_spec_append2):
        buffer = StringIO()
        sql_writer = SQLFormatter(
            specification=fixture_spec_append2)
        sql_writer.format(dataframe=replace_dataframe,
                          path_or_buffer=buffer)

        assert "INSERT INTO".lower() in buffer.getvalue().lower()



import json
import pytest

from mock import mock_open, patch
from os import remove
from pandas import DataFrame

from src.generators.handler import BaseHandler
from src.tests.generators.handler_fixtures import *  # noqa: F403, F401


class TestBaseHundler(object):

    def test_simple_specification_dataframe(self,
                                            mocker,
                                            simple_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = simple_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(specification,
                                               simple_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == simple_specification['size']
        for field in simple_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_argumented_specification_dataframe(self,
                                                mocker,
                                                argumented_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = argumented_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(
            specification,
            argumented_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == argumented_specification['size']
        for field in argumented_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_integer_dataframe(self, mocker, integer_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = integer_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(specification,
                                               integer_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == integer_specification['size']
        for field in integer_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_bool_dataframe(self, mocker, bool_specification):
        with patch('builtins.open',
                   mock_open(read_data=json.dumps(bool_specification))):
            dataset = bool_specification['datasets']['sample']
            expected_size = dataset['size']
            expected_fields = dataset['fields']

            handler = BaseHandler()
            dataframe = handler.generate_dataframe(dataset, dataset['size'])

            assert isinstance(dataframe, DataFrame) is True
            assert dataframe.shape[0] == expected_size
            for field in expected_fields:
                assert dataframe[field['name']].dtype.name == field['expected']

    def test_char_dataframe(self, mocker, char_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = char_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(specification,
                                               char_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == char_specification['size']
        for field in char_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_float_dataframe(self, mocker, float_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = float_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(specification,
                                               float_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == float_specification['size']
        for field in float_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_integer_sequence_dataframe(self,
                                        mocker,
                                        integer_sequence_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = integer_sequence_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(
            specification,
            integer_sequence_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == integer_sequence_specification['size']
        for field in integer_sequence_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_timestamp_sequence_dataframe(self,
                                          mocker,
                                          timestamp_sequence_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = timestamp_sequence_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(
            specification,
            timestamp_sequence_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == timestamp_sequence_specification['size']
        for field in timestamp_sequence_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_string_dataframe(self, mocker, string_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = string_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(specification,
                                               string_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == string_specification['size']
        for field in string_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_timestamp_dataframe(self, mocker, timestamp_specification):
        mock = mocker.patch \
                     .object(BaseHandler, 'valid_specification')
        mock.return_value = timestamp_specification
        handler = BaseHandler()
        specification = handler.valid_specification()
        dataframe = handler.generate_dataframe(specification,
                                               timestamp_specification['size'])

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == timestamp_specification['size']
        for field in timestamp_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_basehandler_valid(self, valid_specification):
        handler = BaseHandler()

        with open('valid_spec.json', 'w') as f:
            json.dump(valid_specification, f)

        config = handler.valid_specification('valid_spec.json')

        assert valid_specification == config

        remove('valid_spec.json')

    def test_basehandler_no_dataset(self, no_datasets_specification):
        handler = BaseHandler()
        with open('invalid_spec.json', 'w') as f:
            json.dump(no_datasets_specification, f)

        with pytest.raises(ValueError):
            handler.valid_specification('invalid_spec.json')
        remove('invalid_spec.json')

    def test_basehandler_no_dataset_ids(self, invalid_no_ids_dataset):
        handler = BaseHandler()
        with open('invalid_spec.json', 'w') as f:
            json.dump(invalid_no_ids_dataset, f)

        with pytest.raises(ValueError):
            handler.valid_specification('invalid_spec.json')
        remove('invalid_spec.json')

    def test_basehandler_no_dataset_size(self, invalid_no_size_dataset):
        handler = BaseHandler()
        with open('invalid_spec.json', 'w') as f:
            json.dump(invalid_no_size_dataset, f)

        with pytest.raises(ValueError):
            handler.valid_specification('invalid_spec.json')
        remove('invalid_spec.json')

    def test_basehandler_no_dataset_locale(self, invalid_no_locale_dataset):
        handler = BaseHandler()
        with open('invalid_spec.json', 'w') as f:
            json.dump(invalid_no_locale_dataset, f)

        with pytest.raises(ValueError):
            handler.valid_specification('invalid_spec.json')
        remove('invalid_spec.json')

    def test_date_format_invalid_specification(self,
                                               mocker,
                                               invalid_dateformat):
        with patch('builtins.open',
                   mock_open(read_data=json.dumps(invalid_dateformat))):
            handler = BaseHandler()
            with pytest.raises(ValueError):
                handler.valid_specification('')

    def test_malformed_json(self, mocker, malformed_json):
        with patch('builtins.open',
                   mock_open(read_data=malformed_json)):
            handler = BaseHandler()
            with pytest.raises(ValueError):
                handler.valid_specification('')

    def test_unknown_type(self, mocker, unknown_type_spec):
        with patch('builtins.open',
                   mock_open(read_data=json.dumps(unknown_type_spec))):
            handler = BaseHandler()
            with pytest.raises(ValueError):
                handler.valid_specification('')

    def test_basehandler_valid_replace(self, valid_spec_for_replace_rules):
        handler = BaseHandler()

        with open('valid_spec_for_replace_rules.json', 'w') as f:
            json.dump(valid_spec_for_replace_rules, f)

        config = handler.valid_specification(
            'valid_spec_for_replace_rules.json')

        assert valid_spec_for_replace_rules == config

        remove('valid_spec_for_replace_rules.json')

    def test_basehandler_invalid_replace_without_schema(
                self,
                invalid_spec_for_replace_rules_without_schema):
        handler = BaseHandler()

        with open(
             'invalid_spec_for_replace_rules_without_schema.json', 'w') as f:
            json.dump(invalid_spec_for_replace_rules_without_schema, f)
        with pytest.raises(ValueError):
            handler.valid_specification(
             'invalid_spec_for_replace_rules_without_schema.json')

        remove('invalid_spec_for_replace_rules_without_schema.json')

    def test_basehandler_invalid_replace_without_sqltype(
                self,
                invalid_spec_for_replace_rules_without_sqltype):
        handler = BaseHandler()

        with open(
             'invalid_spec_for_replace_rules_without_sqltype.json', 'w') as f:
            json.dump(invalid_spec_for_replace_rules_without_sqltype, f)
        with pytest.raises(ValueError):
            handler.valid_specification(
             'invalid_spec_for_replace_rules_without_sqltype.json')

        remove('invalid_spec_for_replace_rules_without_sqltype.json')



from pytest import raises

from src.lib.postgresql.psql import PostgresSqlPsql
from src.lib.shell import ShellError
from src.lib.shell import Shell


class TestPostgreSqlPsql(object):
    def test_execute_success(self, mocker):
        query = ''
        expected = 0, b'', None

        mock = mocker.patch.object(Shell, 'execute_command')
        mock.return_value = expected

        psql = PostgresSqlPsql(host='',
                               port='',
                               database='',
                               user='',
                               password='')
        query_result = psql.execute_query(query)

        mock.assert_called()
        assert query_result == expected[1].decode()

    def test_execute_failure(self, mocker):
        query = ''
        expected = 1, b'', b'Error'

        mock = mocker.patch.object(Shell, 'execute_command')
        mock.return_value = expected

        psql = PostgresSqlPsql(host='',
                               port='',
                               database='',
                               user='',
                               password='')
        with raises(ShellError):
            psql.execute_query(query)

        mock.assert_called()



from pandas import DataFrame
import gcsfs


class GCSRemoteWriter(object):
    key = 'gcs'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification
        self.gcstorage_fs = gcsfs.GCSFileSystem(token='google_default')

    def write(self, dataframe: DataFrame) -> None:
        if self.gcstorage_fs.token is None:
            raise Exception("Without permission, please login")
        options = self.specification['options']
        key = f'gs://{options["bucket"]}/{options["key"]}'
        file_buffer = self.gcstorage_fs.open(key, 'w')
        self.formatter.format(dataframe=dataframe,
                              path_or_buffer=file_buffer)
        return key

    @staticmethod
    def rules():
        return {'required': {'options.bucket': {'none': False, 'type': str},
                             'options.key': {'none': False, 'type': str}},
                'optional': {}}



import json
import requests

from mock import mock_open, patch
from pytest import raises

from src.lib.formatters.csv import CsvFormatter
from src.lib.writers.remotes.s3_presigned_url import S3PresignedUrlRemoteWriter
from src.tests.lib.writers.remotes.fixtures import *  # noqa: F403, F401


class TestS3PresignedUrlRemoteWriter(object):
    """Unit-test for S3PresignedUrlRemoteWriter class"""

    def test_writting_csv_with_records(self,
                                       mocker,
                                       specification_with_url,
                                       pandas_dataframe_with_data,
                                       signed_post_file):
        mock_post = mocker.patch.object(requests, 'post')
        mock_post.return_value.raise_for_status = lambda: None

        with patch('builtins.open',
                   mock_open(read_data=signed_post_file)) as mock_file:
            formatter = CsvFormatter(specification={})
            writer = S3PresignedUrlRemoteWriter(
                formatter=formatter,
                specification=specification_with_url
            )
            signed_url = writer.write(dataframe=pandas_dataframe_with_data)

            mock_post.assert_called()
            mock_file.assert_called()

            expected = json.loads(signed_post_file)['url']

            assert signed_url == expected

    def test_writting_csv_without_records(self,
                                          mocker,
                                          pandas_dataframe_without_data,
                                          specification_with_url,
                                          signed_post_file):
        mock_post = mocker.patch.object(requests, 'post')
        mock_post.return_value.raise_for_status = lambda: None

        with patch('builtins.open',
                   mock_open(read_data=signed_post_file)) as mock_file:
            formatter = CsvFormatter(specification={})
            writer = S3PresignedUrlRemoteWriter(
                formatter=formatter,
                specification=specification_with_url
            )
            signed_url = writer.write(dataframe=pandas_dataframe_without_data)

            mock_post.assert_called()
            mock_file.assert_called()

            expected = json.loads(signed_post_file)['url']

            assert signed_url == expected

    def test_without_url_and_valid_input(self,
                                         mocker,
                                         pandas_dataframe_without_data,
                                         specification_without_url,
                                         signed_post_file):
        expected = json.loads(signed_post_file)['url']
        mock_post = mocker.patch.object(requests, 'post')
        mock_post.return_value.raise_for_status = lambda: None
        mock_input = mocker.patch('builtins.input')
        mock_input.return_value = expected
        mock_input = mocker.patch('os.path.isfile')
        mock_input.return_value = True

        with patch('builtins.open',
                   mock_open(read_data=signed_post_file)) as mock_file:
            formatter = CsvFormatter(specification={})
            writer = S3PresignedUrlRemoteWriter(
                formatter=formatter,
                specification=specification_without_url
            )
            writer.before_write()
            signed_url = writer.write(dataframe=pandas_dataframe_without_data)

            mock_post.assert_called()
            mock_file.assert_called()

            assert signed_url == expected

    def test_without_url_and_invalid_input(self,
                                           mocker,
                                           pandas_dataframe_without_data,
                                           specification_without_url,
                                           signed_post_file):
        expected = json.loads(signed_post_file)['url']
        mock_post = mocker.patch.object(requests, 'post')
        mock_post.return_value.raise_for_status = lambda: None
        mock_input = mocker.patch('builtins.input')
        mock_input.return_value = expected
        mock_input = mocker.patch('os.path.isfile')
        mock_input.return_value = False

        with patch('builtins.open',
                   mock_open(read_data=signed_post_file)) as mock_file:
            formatter = CsvFormatter(specification={})
            writer = S3PresignedUrlRemoteWriter(
                formatter=formatter,
                specification=specification_without_url
            )

            with raises(ValueError):
                writer.before_write()

                mock_post.assert_called()
                mock_file.assert_called()



from pytest import fixture


def pandas_dataframe(data):
    from pandas import DataFrame

    dataframe = DataFrame(data=data)

    return dataframe


@fixture
def pandas_dataframe_with_data():
    data = [{"Column1": "Value_1", "Column2": "Value_10"},
            {"Column1": "Value_2", "Column2": "Value_11"},
            {"Column1": "Value_3", "Column2": "Value_12"},
            {"Column1": "Value_4", "Column2": "Value_13"},
            {"Column1": "Value_5", "Column2": "Value_14"},
            {"Column1": "Value_6", "Column2": "Value_15"}]
    return pandas_dataframe(data=data)


@fixture
def pandas_dataframe_without_data():
    return pandas_dataframe(data=[])


@fixture
def specification():
    from os.path import abspath
    from uuid import uuid4

    absolute_path = abspath(".")
    file_name = uuid4()

    return {'uri': f"{absolute_path}/{file_name}"}


@fixture
def sql_specification_format():
    def _spec(table_name, index=False, index_label=None):
        return {
            'options': {
                'table_name': table_name,
                'batch_size': 2,
                'schema': {
                    'Column1': {
                        'quoted': True,
                        'sqltype': "VARCHAR(50)"
                    },
                    'Column2': {
                        'quoted': True,
                        'sqltype': "VARCHAR(50)"
                    },
                },
                'mode': 'replace',
                'index': index,
                'index_label': index_label
            }
        }
    return _spec


@fixture
def sql_specification_writer():
    return {
        'options': {
            'engine': 'mysql',
            'method': 'direct',
            'host': '172.17.0.2',
            'port': 3306,
            'database': 'mydb',
            'username': 'root'
        }
    }



from io import StringIO

from src.lib.formatters.json import JsonFormatter
from src.tests.lib.formatters.fixtures import *  # noqa: F403, F401


class TestJsonFormatter(object):
    """Unit-test of JsonWriter class"""

    def test_writting_dataframe_with_records(self,
                                             pandas_dataframe_with_data):
        buffer = StringIO()
        json_writer = JsonFormatter(specification={})
        json_writer.format(dataframe=pandas_dataframe_with_data,
                           path_or_buffer=buffer)

        assert isinstance(buffer.getvalue(), str) is True
        assert len(buffer.getvalue()) > 0

    def test_writting_dataframe_without_records(self,
                                                pandas_dataframe_without_data):
        buffer = StringIO()
        json_writer = JsonFormatter(specification={})
        json_writer.format(dataframe=pandas_dataframe_without_data,
                           path_or_buffer=buffer)

        assert isinstance(buffer.getvalue(), str) is True
        assert len(buffer.getvalue()) == 0



from src.generators.datatypes import Metadata


class FieldsConfiguration(object):

    @staticmethod
    def get_rules(field, locale):
        type_ = field.get('type')
        metadata = Metadata(locale)
        generators = metadata.get_generators_map()

        if type_ not in generators:
            raise ValueError(f"Field type `{type_}` not found")

        type_class = generators[type_]['type']

        return type_class.rules()



import requests

from io import StringIO
from pandas import DataFrame
from urllib.parse import urlparse


class GCSPresignedUrlRemoteWriter(object):
    key = 'gcs-url'
    signed_url = None

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification

    def before_write(self):
        self.signed_url = self.specification.get('uri')

        if not self.signed_url:
            self.signed_url = input("Enter the signed url: ")

    def write(self, dataframe: DataFrame):
        buffer = StringIO()

        self.formatter.format(dataframe=dataframe,
                              path_or_buffer=buffer)

        try:
            response = requests.put(self.signed_url,
                                    data=buffer.getvalue())
            response.raise_for_status()
        except Exception as e:
            raise requests.RequestException("Can't send data to this given"
                                            "URI try to check if still valid",
                                            e)
        return self.__get_file_path_from_resp(response)

    def __get_file_path_from_resp(self, resp):
        return f"gs://{urlparse(resp.url).path}"

    @staticmethod
    def rules():
        return {'required': {},
                'optional': {'uri': {'none': False, 'type': str}}}



import platform
from os import getenv

from src.lib.shell import Shell

MYSQL_CLI_BINPATH = getenv("MYSQL_CLI_BINPATH")


class MySQLConnection(object):
    connection = ("--host={host} --port={port} --user={user} "
                  "--password={password}")

    def __init__(self, host, username, password, database, port=3306,
                 **kwargs):
        self.shell = Shell()
        self.base_connection = self.connection.format(
            host=host,
            port=port,
            user=username,
            password=password,
        )
        self.base_connection = self.base_connection.split(" ")
        self.database = database

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def execute_query(self, query, *args, **kwargs):
        command = [
            MYSQL_CLI_BINPATH or 'mysql',
        ]

        if platform.system() == "Windows":
            command.append("--default-character-set=utf8")

        for argoption in self.base_connection:
            command.append(argoption)
        command.append(self.database)
        command.append('-e')
        command.append(query)

        self.shell.execute(command=command)

    def close(self):
        pass



import json

from alive_progress import alive_bar
from pandas import DataFrame, Series

from src.lib.config import general_rules, datetime_rules
from src.lib.config.fields import FieldsConfiguration
from src.lib.config.formatters import FormattersConfiguration
from src.lib.config.inspector import ConfigurationInpector
from src.lib.config.serializers import SerializersConfiguration
from src.generators.datatypes import Metadata


class BaseHandler(object):
    def __init__(self):
        self.inspector = ConfigurationInpector()

    def __validate(self, clazz, iterable, trace, **kwargs):
        for item in iterable:
            try:
                rules = clazz.get_rules(item, **kwargs)
            except ValueError as err:
                type_ = item.get('type', '')
                name = item.get('name', type_)
                message = (f"Key {trace}{name} ({err})")
                raise ValueError(message)
            self.inspector.inspect_rules(rules=rules['required'],
                                         configuration=item)
            self.inspector.inspect_rules(rules=rules['optional'],
                                         configuration=item,
                                         optional=True)

    def __validate_fields(self, dataset, trace):
        fields = dataset['fields']
        self.__validate(clazz=FieldsConfiguration,
                        iterable=fields,
                        locale=dataset['locale'],
                        trace=f"{trace}fields.")

    def __validate_format(self, dataset, trace):
        self.__validate(clazz=FormattersConfiguration,
                        iterable=[dataset['format']],
                        trace=f"{trace}format.")

    def __validate_serializers(self, dataset, trace):
        serializers = dataset['serializers']['to']
        self.__validate(clazz=SerializersConfiguration,
                        iterable=serializers,
                        trace=f"{trace}serializers.to.")

    def valid_specification(self, file_path):
        with open(file_path, 'r') as file:
            try:
                configuration = json.loads(file.read())
            except Exception:
                raise ValueError("Malformed specification file/blank field")

        self.inspector.inspect_rules(rules=general_rules,
                                     configuration=configuration)

        for dataset_key in configuration['datasets']:
            dataset = configuration['datasets'][dataset_key]

            trace = f"datasets.{dataset_key}."
            self.__validate_fields(dataset, trace)
            self.__validate_format(dataset, trace)
            self.__validate_serializers(dataset, trace)

        for dataset_key in configuration['datasets']:
            has_datetime = False
            dataset = configuration['datasets'][dataset_key]
            for field in dataset['fields']:
                if field.get('type') == 'date_time':
                    has_datetime = True
                    break
            if has_datetime:
                self.inspector \
                    .inspect_rules(rules=datetime_rules,
                                   configuration=dataset)

        return configuration

    def generate_dataframe(self, specification: dict, size) -> DataFrame:
        fields = specification['fields']
        locale = specification['locale']
        generators_map = Metadata(locale=locale).get_generators_map()

        dataframe = DataFrame()
        for field in fields:
            field_name = field['name']
            field_type = field['type']
            field_arguments = field['generator']

            generator_class = generators_map[field_type]['type']
            generator = generator_class(**field_arguments)

            with alive_bar(size) as bar:
                # E731 do not assign a lambda expression, use a def
                def progress(): bar(text=f"ggenerating {field_name}")

                series = Series(generator.generate_records(num_of_records=size,
                                                           progress=progress))

            dataframe[field_name] = series.values

        return dataframe



from src.lib.formatters.csv import CsvFormatter
from src.lib.formatters.json import JsonFormatter
from src.lib.formatters.sql import SQLFormatter

formatters = {
    CsvFormatter.key: CsvFormatter,
    JsonFormatter.key: JsonFormatter,
    SQLFormatter.key: SQLFormatter
}



from src.lib.writers.remotes.gcs_presigned_url \
    import GCSPresignedUrlRemoteWriter
from src.lib.formatters.csv import CsvFormatter
from src.tests.lib.writers.remotes.fixtures import *  # noqa: F403, F401
import pytest
import requests


class TestGcsWriter:
    def test_writting_csv_with_records(self,
                                       mocker,
                                       specification_url_gcs,
                                       pandas_dataframe_with_data):

        mock = mocker.patch.object(requests, 'put')
        resp = requests.Response()
        resp.status_code = 200
        mock.return_value = resp

        formatter = CsvFormatter(specification={})
        writer = \
            GCSPresignedUrlRemoteWriter(formatter=formatter,
                                        specification=specification_url_gcs)
        writer.before_write()
        writer.write(dataframe=pandas_dataframe_with_data)
        mock.assert_called()

    def test_writting_csv_without_records(self,
                                          mocker,
                                          pandas_dataframe_without_data,
                                          specification_url_gcs):

        mock = mocker.patch.object(requests, 'put')
        resp = requests.Response()
        resp.status_code = 200
        mock.return_value = resp
        formatter = CsvFormatter(specification={})
        writer = \
            GCSPresignedUrlRemoteWriter(formatter=formatter,
                                        specification=specification_url_gcs)
        writer.before_write()
        writer.write(dataframe=pandas_dataframe_without_data)
        mock.assert_called()

    def test_writting_raised(self,
                             mocker,
                             pandas_dataframe_without_data,
                             specification_url_gcs):

        formatter = CsvFormatter(specification={})
        writer = \
            GCSPresignedUrlRemoteWriter(formatter=formatter,
                                        specification=specification_url_gcs)

        with pytest.raises(requests.RequestException):
            writer.before_write()
            writer.write(dataframe=pandas_dataframe_without_data)

    def test_without_url_and_valid_input(self,
                                         mocker,
                                         pandas_dataframe_without_data,
                                         gcs_without_url,
                                         specification_url_gcs):
        expected = specification_url_gcs['uri']
        response = requests.Response()
        response.status_code = 200
        response.url = expected

        mock_put = mocker.patch.object(requests, 'put')
        mock_put.return_value = response

        mock_input = mocker.patch('builtins.input')
        mock_input.return_value = expected

        formatter = CsvFormatter(specification={})
        writer = \
            GCSPresignedUrlRemoteWriter(formatter=formatter,
                                        specification=gcs_without_url)
        writer.before_write()
        signed_url = writer.write(dataframe=pandas_dataframe_without_data)

        mock_put.assert_called()

        assert signed_url == "gs://"



from pandas import DataFrame
from sqlalchemy import create_engine
from src.lib.writers.databases import DatabaseWriter
from src.lib.mysql.mysql import MySQLConnection


class MysqlDatabaseWriter(DatabaseWriter):
    @staticmethod
    def rules():
        return {
            'required': {
                'options.engine': {'none': False, 'type': str},
                'options.host': {'none': False, 'type': str},
                'options.database': {'none': False, 'type': str},
                'options.username': {'none': False, 'type': str}
            },
            'optional': {
                'options.method': {'none': False, 'type': str},
                'options.port': {'none': False, 'type': int},
                'options.schema': {'none': False, 'type': str}
            }
        }


class MysqlClientDatabaseWriter(MysqlDatabaseWriter):
    key = 'sql'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification

    def write(self, dataframe: DataFrame):
        parameters = self.specification \
                         .get('options')

        sql = self.formatter \
                  .format(dataframe=dataframe,
                          path_or_buffer=None)
        if sql is not None:
            with MySQLConnection(**parameters) as connection:
                for query in sql.split(';'):
                    connection.execute_query(query)

        return "Inserted with success"


class MysqlDirectDatabaseWriter(MysqlDatabaseWriter):
    """Class that receive pandas dataframe
    and write it on a SGDB
    """
    key = 'sql'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.default = {
            'schema': None
        }
        self.specification = specification

    def engine(self, params):
        return create_engine('mysql+mysqlconnector://'
                             f"{params['username']}:"
                             f"{params['password']}@"
                             f"{params['host']}:"
                             f"{params.get('port', 3306)}/"
                             f"{params['database']}")

    def write(self, dataframe: DataFrame) -> None:
        """Write all dataframe on a DataBase Table.

        Keyword arguments:
            - dataframe - pandas.Dataframe: dataframe containing the records
        """
        parameters = self.default
        options = self.specification.get('options', {})
        parameters.update(options)

        connection = self.engine(parameters)

        self.formatter.format(dataframe=dataframe,
                              path_or_buffer=connection,
                              method='direct',
                              schema=parameters['schema'])
        return "Inserted with success"



from pandas import DataFrame


class Sql(object):
    def __parse_rows(self, dataframe: DataFrame, schema: dict = {}) -> list:
        columns = dataframe.columns
        values = []
        for index, row in dataframe.iterrows():
            values.append(self.__parse_row(row, columns, schema))
        return values

    def __parse_row(self, row, columns, schema: dict = {}):
        row_value_sql = "("
        first_column = True
        for column in columns:
            if not first_column:
                row_value_sql += ", "
            first_column = False
            if self.is_quoted(column, schema, row[column]):
                row_value_sql += "'" + str(row[column]) + "'"
            else:
                row_value_sql += str(row[column])
        row_value_sql += ")"
        return row_value_sql

    def is_quoted(self, column, schema, value):
        return self.is_text(value) or \
            column in schema and \
            schema.get(column, dict()).get('quoted', False) is True

    def is_text(self, value):
        return isinstance(value, str)

    def insert_statement(self, dataframe: DataFrame,
                         table_name: str,
                         schema: dict):
        columns = dataframe.columns
        values = self.__parse_rows(dataframe, schema)
        return f"INSERT INTO {table_name} " \
               f"({', '.join(columns)}) \n"\
               "VALUES\n" + ',\n'.join(values)

    def append_statement(self, dataframe, params):
        table_name = params.get("table_name")
        schema = params.get("schema", {})
        batch_size = params['batch_size']
        query = ""
        for index in range(0, dataframe.shape[0], batch_size):
            query += \
                self.insert_statement(dataframe[index:index+batch_size],
                                      table_name,
                                      schema)
            query += ";\n\n"
        return query

    def create_append_statement(self, dataframe, params):
        mode = params.get('mode')
        if mode == 'truncate' or mode == 'append':
            query = self.append_statement(dataframe, params)
        else:
            query = self.create_table_statement(params)
            query += self.append_statement(dataframe, params)
        return query

    def truncate_statement(self, dataframe, params):
        table_name = params.get("table_name")
        query = f"TRUNCATE {table_name};\n\n"
        query += self.append_statement(dataframe, params)
        return query

    def replace_statement(self, dataframe, params):
        table_name = params.get('table_name')
        replace = f"DROP TABLE IF EXISTS {table_name};\n"
        replace += self.create_table_statement(params)
        append = self.append_statement(dataframe, params)
        return f"{replace} \n{append}"

    def create_table_statement(self, params):
        schema = params.get('schema')
        table_name = params.get('table_name')
        fields = "("
        cont = 0
        if params.get("index"):
            fields += params.get("index_label") + " INTEGER,"
        for new_field in schema:
            sql_type = schema[new_field].get('sqltype')
            fields += new_field + " " + sql_type
            cont += 1
            if cont < len(schema):
                fields += ", "
        fields += ");"
        return f"CREATE TABLE IF NOT EXISTS {table_name} {fields}\n\n"



from src.tests.lib.formatters.fixtures import pandas_dataframe
from pytest import fixture


@fixture
def mysql_specification():
    return {
        "type": "sql",
        "options": {
            "engine": "mysql",
            "method": "cli",
            "host": "",
            "port": 3306,
            "database": "",
            "schema": "",
            "username": "",
            "password": ""
        }
    }


@fixture
def mysql_specification_script():
    return {
        "type": "file",
        "uri": "path/to/file.sql"
    }


@fixture
def postgres_specification_cli():
    return {
        "type": "sql",
        "options": {
            "engine": "postgres",
            "method": "cli",
            "host": "",
            "port": 5432,
            "database": "",
            "username": "",
            "password": ""
        }
    }


@fixture
def postgres_specification_direct():
    return {
        "type": "sql",
        "options": {
            "engine": "postgres",
            "method": "direct",
            "host": "",
            "port": 5432,
            "schema": None,
            "database": "",
            "username": "",
            "password": ""
        }
    }


@fixture
def dataframe():
    data = [
        {
            "id": 1,
            "name": "Sophie Silveira",
            "age": 65,
        },
        {
            "id": 2,
            "name": "Davi Lucca Duarte",
            "age": 105,
        },
        {
            "id": 3,
            "name": "Luna da Rosa",
            "age": 56,
        },
        {
            "id": 4,
            "name": "Vitor Melo",
            "age": 74,
        },
        {
            "id": 5,
            "name": "Catarina Correia",
            "age": 6,
        }
    ]
    return pandas_dataframe(data=data)



import os
from os import getenv

from setuptools import setup, find_packages

VERSION = getenv("TRAVIS_TAG", 'v0.0')


def get_version(version):
    if version[0] == 'v':
        return version[1:]
    if version[:4] == 'dev-':
        return version[4:]
    raise ValueError("Malformed tag version")


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="ggenerator",
    author="Datenworks",
    author_email="contato@datenworks.com",
    description=("A tool capable to generate fake data with a "
                 "given specification defined as a JSON DSL"),
    url="https://datenworks.com/",
    license='MIT',
    project_urls={
        "Source Code": "https://github.com/Datenworks/ggenerator"
    },
    py_modules=['main'],
    long_description=read("README.md"),
    version=get_version(VERSION),
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "python-dateutil>=2.8.1",
        "pandas>=1.0.1",
        "click>=7.1.1",
        "tabulate>=0.8.6",
        "requests>=2.23.0",
        "gcsfs>=0.6.0",
        "s3fs>=0.4.0",
        "faker>=4.0.1",
        "sqlalchemy>=1.3.16",
        "mysql-connector-python>=8.0.19",
        "alive-progress>=1.5.1",
        "azure-storage-blob>=12.3.0",
        "psycopg2-binary>=2.8.5"
    ],
    extras_require={
        "mysql": ['mysqlclient>=1.4.6']
    },
    entry_points="""
        [console_scripts]
        ggenerator=main:run
    """,
    )



from src.lib.writers.file import FileWriter
from src.lib.writers.remotes.s3 import S3RemoteWriter
from src.lib.writers.remotes.s3_presigned_url import S3PresignedUrlRemoteWriter
from src.lib.writers.remotes.gcs import GCSRemoteWriter
from src.lib.writers.remotes.azure_bs import AzureBSRemoteWriter
from src.lib.writers.\
    remotes.gcs_presigned_url import GCSPresignedUrlRemoteWriter
from src.lib.writers.databases.mysql import MysqlClientDatabaseWriter, \
    MysqlDirectDatabaseWriter
from src.lib.writers.databases.postgresql import PostgresDirectDatabaseWriter,\
    PostgreSqlClientDatabaseWriter

writers = {
    S3PresignedUrlRemoteWriter.key: S3PresignedUrlRemoteWriter,
    S3RemoteWriter.key: S3RemoteWriter,
    GCSRemoteWriter.key: GCSRemoteWriter,
    AzureBSRemoteWriter.key: AzureBSRemoteWriter
}

uri_writers = {
    FileWriter.key: FileWriter,
    GCSPresignedUrlRemoteWriter.key: GCSPresignedUrlRemoteWriter
}

database_writers = {
    'mysql-cli': MysqlClientDatabaseWriter,
    'mysql-direct': MysqlDirectDatabaseWriter,
    'postgres-cli': PostgreSqlClientDatabaseWriter,
    'postgres-direct': PostgresDirectDatabaseWriter
}

writers.update(uri_writers)
writers.update(database_writers)



from pandas import DataFrame

from tabulate import tabulate
from src.generators.basehandler import BaseHandler


class DryRunHandler(object):
    """DryRunHandler is responsible to give
    a dataframe preview.
    """
    def __init__(self, arguments: dict):
        self.file_path = arguments['config_file']
        self.base = BaseHandler()
        self.specification = self.valid_specification_dryrun()

    def valid_specification_dryrun(self):
        valid = self.base.valid_specification(self.file_path)
        return valid

    def generate(self):
        datasets = self.specification.get('datasets')
        for key in datasets.keys():
            dataset = datasets[key]
            dataset_fields = dataset['fields']
            dataframe = self.generate_dryrun(dataset)
            dryrun = self.print_dryrun(dataframe, key, dataset_fields)
        return dryrun

    def print_dryrun(self, dataframe, key, dataset_fields):
        print("Dataset: ", key)
        self.__print_fields(dataset_fields)
        print("-------------------")
        print("First 10 rows: \n")
        self.__print_dataframe(dataframe)
        print("-------------------")

    def __print_fields(self, dataset_fields):
        for field in dataset_fields:
            print("   ", field['name'], ":", field['type'])

    def __print_dataframe(self, dataframe):
        print(tabulate(
            dataframe, tablefmt="fancy_grid",
            headers=dataframe.columns))

    def generate_dryrun(self, specification: dict) -> DataFrame:
        dataframe = self.base.generate_dataframe(specification, 10)
        return dataframe



from csv import reader
from os import remove
from os.path import exists, isfile

from src.lib.formatters.csv import CsvFormatter
from src.lib.writers.file import FileWriter
from src.tests.lib.writers.fixtures import *  # noqa: F403, F401


class TestFileWriter(object):
    """Unit-test for FileWriter class"""

    def test_writting_csv_with_records(self,
                                       pandas_dataframe_with_data,
                                       specification):
        formatter = CsvFormatter(specification={})
        writer = FileWriter(formatter=formatter, specification=specification)
        writer.write(dataframe=pandas_dataframe_with_data)

        assert exists(specification['uri']) is True
        assert isfile(specification['uri']) is True

        with open(specification['uri']) as csv_file:
            csv_reader = reader(csv_file, delimiter=',')
            csv_reader = [x for x in csv_reader]
            headers = csv_reader[0]
            rows = csv_reader[1:len(csv_reader)]
            csv_reader = {headers[index]: [row[index]
                                           for row in rows]
                          for index in range(0, len(headers))}
            expected = pandas_dataframe_with_data.to_dict(orient="list")

            assert csv_reader == expected

        remove(specification['uri'])
        assert exists(specification['uri']) is False

    def test_writting_dataframe_without_records(self,
                                                pandas_dataframe_without_data,
                                                specification):
        formatter = CsvFormatter(specification={})
        writer = FileWriter(formatter=formatter, specification=specification)
        writer.write(dataframe=pandas_dataframe_without_data)

        assert exists(specification['uri']) is False
        assert isfile(specification['uri']) is False




def args_to_kwargs(params: list, separator: str) -> dict:
    kwargs = {}
    for param in params:
        assert param.count('=') == 1
        key, value = param.split('=')
        value = eval(value) if is_evaluable(value) else value
        kwargs.update({key: value})
    return kwargs


def is_evaluable(value):
    try:
        eval(value)
        return True
    except Exception:
        return False



from pytest import raises

from src.lib.config.formatters import FormattersConfiguration
from src.tests.lib.config.fixtures import *  # noqa: F403, F401


class TestFormattersConfiguration(object):
    """Unit-test of TestFormattersConfiguration class"""

    def test_csv(self, csv_format_sample):
        rules = FormattersConfiguration.get_rules(csv_format_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_json(self, json_format_sample):
        rules = FormattersConfiguration.get_rules(json_format_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_sql(self, sql_format_sample):
        rules = FormattersConfiguration.get_rules(sql_format_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_unknown(self, unknown_format_sample):
        with raises(ValueError):
            FormattersConfiguration.get_rules(unknown_format_sample)
